/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <infiniband/verbs.h>
#include <infiniband/mlx5_api.h>
#include <libflexio/flexio.h>

#include <doca_log.h>

DOCA_LOG_REGISTER(FLEXIO_RPC::SAMPLE);

#define IBV_DEVICE_NAME "mlx5_0"			/* Use mlx5_0 device */

/* FlexIO application, generated by DPACC stub */
extern struct flexio_app *flexio_rpc_device;

/* RPC function to be called on the device */
extern flexio_func_t flexio_rpc_calculate_sum;

/*
 * Run the FlexIO RPC sample
 *
 * @argc [in]: Number of arguments
 * @argv [in]: Arguments
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
doca_error_t
flexio_rpc(int argc, char **argv)
{
	(void)argc;
	(void)argv;

	struct ibv_device **dev_list = NULL;		/* List of IB devices */
	struct ibv_device *dev = NULL;			/* Device to use */
	struct ibv_context *dev_ctx = NULL;		/* Device context */

	struct flexio_process *flexio_process;		/* FlexIO process */

	flexio_status ret = FLEXIO_STATUS_SUCCESS;					/* Return value */
	int ibv_list_idx = 0;				/* Index into device list */
	uint64_t rpc_call_result;			/* RPC call result */
	doca_error_t result = DOCA_SUCCESS;

	/* Get list of available IB devices */
	dev_list = ibv_get_device_list(NULL);

	if (dev_list == NULL) {
		DOCA_LOG_ERR("Failed to get IB device list");
		return DOCA_ERROR_DRIVER;
	}

	/* Find the device we want to use */
	for (ibv_list_idx = 0; dev_list[ibv_list_idx] != NULL; ibv_list_idx++) {
		if (strcmp(dev_list[ibv_list_idx]->name, IBV_DEVICE_NAME) == 0) {
			dev = dev_list[ibv_list_idx];
			break;
		}
	}

	if (dev == NULL) {
		DOCA_LOG_ERR("Failed to find device %s", IBV_DEVICE_NAME);
		result = DOCA_ERROR_NOT_FOUND;
		goto devlist_cleanup;
	}

	/* Create device context */
	dev_ctx = ibv_open_device(dev);
	if (dev_ctx == NULL) {
		DOCA_LOG_ERR("Failed to open device %s", IBV_DEVICE_NAME);
		result = DOCA_ERROR_DRIVER;
		goto devlist_cleanup;
	}

	/* Create FlexIO process */
	ret = flexio_process_create(dev_ctx, flexio_rpc_device, NULL, &flexio_process);
	if (ret != FLEXIO_STATUS_SUCCESS) {
		DOCA_LOG_ERR("Failed to create FlexIO process");
		result = DOCA_ERROR_DRIVER;
		goto dev_cleanup;
	}

	/* Call the RPC function */
	ret = flexio_process_call(flexio_process, &flexio_rpc_calculate_sum, &rpc_call_result, 0xDBA0, 16, 0xBABE);
	if (ret != FLEXIO_STATUS_SUCCESS) {
		DOCA_LOG_ERR("Failed to call RPC function");
		result = DOCA_ERROR_DRIVER;
		goto rpc_cleanup;
	}

	/* Print the return value */
	DOCA_LOG_INFO("RPC call result: %lx", rpc_call_result);

rpc_cleanup:
	if (flexio_process)
		flexio_process_destroy(flexio_process);
dev_cleanup:
	if (dev_ctx)
		ibv_close_device(dev_ctx);
devlist_cleanup:
	if (dev_list)
		ibv_free_device_list(dev_list);
	return result;
}
